require 'fileutils'
require 'rbconfig'

module Enviz_SketchUp_Plugin

    module Enviz_Logging

        # Determine operating system
        def self.os
            @os ||= (
            host_os = RbConfig::CONFIG['host_os']
            case host_os
            when /mswin|msys|mingw|cygwin|bccwin|wince|emc/
                :windows
            when /darwin|mac os/
                :macos
            else
                raise Error, "unknown os: #{host_os.inspect}"
            end
            )
        end

        # Set log directory based on operating system
        LOG_DIRECTORY = if os == :windows
            File.join(ENV['APPDATA'], 'Enviz_SketchUp_Plugin', 'Logs')
        else
            File.join(ENV['HOME'], 'Library', 'Logs', 'Enviz_SketchUp_Plugin')
        end

        class LogProxy
            def initialize(original_stream, log_file)
                @log_array = []
                @original_stream = original_stream
                @log_file = log_file
            end

            def write(*args)
                # Join all arguments into a single string and write it to both the log file and the original stream
                output = args.map(&:to_s).join
                @log_array << output
                @log_file.write(output)
                @original_stream.write(output)
            end

            def log_array
                @log_array.dup  # Return a duplicate to prevent modification from outside
            end

            def close
                @log_file.close
            end
      end

      def self.start_logging
        FileUtils.mkdir_p(LOG_DIRECTORY) unless File.directory?(LOG_DIRECTORY)
        # Create a new log file with the current date and time
        current_time = Time.now.strftime("%Y-%m-%d_%H-%M-%S")
        log_file_path = File.join(LOG_DIRECTORY, "plugin_log_#{current_time}.txt")
        file = File.open(log_file_path, 'a')
        $stdout = LogProxy.new($stdout, file)
        $stderr = LogProxy.new($stderr, file)
      end

      def self.stop_logging
        $stdout.close
        $stderr.close
        $stdout = STDOUT
        $stderr = STDERR
      end

      def self.write_logs_to_file(dir)
        File.open(dir, 'w') do |file|
          file.puts "STDOUT Logs:"
          file.puts $stdout.log_array if $stdout.respond_to?(:log_array)
          file.puts "\nSTDERR Logs:"
          file.puts $stderr.log_array if $stderr.respond_to?(:log_array)
        end
      end

    end
end