require 'sketchup'
require 'tempfile'
require "base64"
Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_small_tools')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
Sketchup::require('enviz_sketchup_plugin/enviz_logging')
Sketchup::require('enviz_sketchup_plugin/teleport_handler')
Sketchup::require('enviz_sketchup_plugin/tool_flipper')

module Enviz_SketchUp_Plugin
    module Teleports

        extend self

        include Enviz_Sketchup_Helpers

        # Preview holder
        @preview_dialog ||= nil

        # Array to hold all teleports
        @teleports_list = []

        def self.teleports_list
            @teleports_list
        end

        # PREVIEW ####################

        def self.preview(order)
            view = Sketchup.active_model.active_view
            original_camera = view.camera
            original_position = original_camera.eye
            original_target = original_camera.target
            original_up = original_camera.up
            original_fov = original_camera.fov
            original_perspective = original_camera.perspective?
          
            puts 'Teleport preview here' if DEBUG_MODE
          
            model = Sketchup.active_model
            entities = model.entities
          
            # Find the entity with the matching TeleportOrder
            entity = entities.find do |ent|
              ent.get_attribute('EnvizAttributes', 'TeleportOrder') == order
            end
          
            if entity
              new_camera_position = entity.transformation.origin.offset(Z_AXIS, entity.bounds.max.z) + Geom::Vector3d.new(0, 0, -6)
              
              local_axis = Geom::Vector3d.new(0, 1, 0)
              forward_vector = entity.transformation * local_axis
              forward_vector = Geom::Vector3d.new(forward_vector.x, forward_vector.y, forward_vector.z) # Reassert as vector
              forward_vector.normalize!
              target_point = new_camera_position
              for i in 1..24
                target_point += forward_vector
              end
              new_camera_position = new_camera_position
              for i in 1..12
                new_camera_position += forward_vector
              end
          
              create_and_show_view(
                entity.name,         # Teleport name
                new_camera_position, # Camera location
                target_point,        # Target point
                Geom::Vector3d.new(0, 0, 1),   # Up vector (optional)
                600,                 # Image width
                350                  # Image height
              )
            else
              puts "No entity found with TeleportOrder #{order}" if DEBUG_MODE
              return
            end
          
            # Restore the original camera settings using individual components
            new_camera = Sketchup::Camera.new(original_position, original_target, original_up)
            new_camera.perspective = original_camera.perspective?
            new_camera.aspect_ratio = original_camera.aspect_ratio if original_camera.perspective?
            new_camera.fov = original_fov
            new_camera.perspective = original_perspective
            view.camera = new_camera
          
            view.refresh

          end

        def create_and_show_view(teleport_name, location, target, up_vector = Z_AXIS, width = 600, height = 350)
            setup_camera(location, target, up_vector)
            base64_image = capture_view_to_base64(width, height)
            show_image_in_html(teleport_name, base64_image)
            
        end

        def setup_camera(location, target, up_vector = Z_AXIS)
            view = Sketchup.active_model.active_view
            new_camera = Sketchup::Camera.new(location, target, up_vector)
            new_camera.perspective = true
            new_camera.fov = 95
            view.camera = new_camera
        end
        def capture_view_to_base64(width, height)
            temp_file = Tempfile.new(['sketchup_capture', '.png'])
            Sketchup.active_model.active_view.write_image(
              filename: temp_file.path,
              width: width-18,
              height: height-50,
              antialias: true,
              transparent: true
            )
          
            # Convert the image to a base64 string
            image_data = File.binread(temp_file.path)
            base64_image = Base64.strict_encode64(image_data)
            temp_file.unlink  # Delete the temp file
          
            "data:image/png;base64,#{base64_image}"
        end

        def show_image_in_html(teleport_name, base64_image)
            html_content = <<-HTML
              <html>
              <body>
                <img src="#{base64_image}" alt="SketchUp View">
              </body>
              </html>
            HTML
        
            # Check if the dialog already exists and is visible
            if @preview_dialog && @preview_dialog.visible?
                @preview_dialog.set_html(html_content)  # Update existing dialog
                @preview_dialog.bring_to_front          # Bring the dialog to the front if it's behind other windows
            else
                # Create a new dialog if it does not exist
                @preview_dialog = UI::HtmlDialog.new({
                  dialog_title: "Enviz Teleport Preview",
                  scrollable: false,
                  resizable: false,
                  width: 600,
                  height: 350,
                  style: UI::HtmlDialog::STYLE_DIALOG
                })
                
                @preview_dialog.set_html(html_content)
                @preview_dialog.show
            end
            @preview_dialog.center
        end
        

        # PREVIEW ####################

        def self.set_teleport_name(text)
            model = Sketchup.active_model
            selection = model.selection
            entity = selection[0]
            
            if selection.length == 1 && entity
                teleport_type = entity.get_attribute('EnvizAttributes', 'TeleportType')
                if teleport_type == 'EnvizTeleport'
                    entity.name = text
                else
                    puts "Selected entity is not a teleport." if DEBUG_MODE
                end
            else
                puts "Please select exactly one entity." if DEBUG_MODE
            end    
        end

        def self.handle_rotation_change(rotation_val)
            model = Sketchup.active_model
            selection = model.selection
            entity = selection[0]
            
            if selection.length == 1 && entity
                teleport_type = entity.get_attribute('EnvizAttributes', 'TeleportType')
                if teleport_type == 'EnvizTeleport'
                    angle_degrees = rotation_val.to_f
                    angle_radians = angle_degrees * Math::PI / 180.0
        
                    # Invert the rotation angle by negating the radians
                    inverted_angle_radians = -angle_radians  # This is the key change
        
                    # Get the current transformation to preserve the translation and scaling
                    current_transformation = entity.transformation
                    origin = current_transformation.origin
        
                    # Create a new rotation transformation around the Z-axis from the origin using the inverted angle
                    new_rotation = Geom::Transformation.rotation(origin, Z_AXIS, inverted_angle_radians)
        
                    # Extract the translation part of the existing transformation
                    translation = Geom::Transformation.translation(current_transformation.origin)
        
                    # Apply translation to the new rotation
                    final_transformation = new_rotation * translation
        
                    model.start_operation('Set Global Rotation', true)
                    entity.transformation = final_transformation
                    model.commit_operation
                else
                    puts "Selected entity is not a teleport." if DEBUG_MODE
                end
            else
                puts "Please select exactly one entity." if DEBUG_MODE
            end
        end

        # Function to build a list of all entities marked as EnvizTeleports
        def self.update_teleports_list
            model = Sketchup.active_model
            entities = model.entities
          
            # Clear the existing list to avoid duplicates
            @teleports_list.clear
          
            # Collect entities with the specific attribute
            teleports = entities.select do |entity|
              if entity.is_a?(Sketchup::ComponentInstance) || entity.is_a?(Sketchup::Group)
                teleport_type = entity.get_attribute('EnvizAttributes', 'TeleportType')
                teleport_type == 'EnvizTeleport'
              else
                false
              end
            end
          
            # Sort the collected entities by their 'TeleportOrder' attribute
            sorted_teleports = teleports.sort_by do |entity|
              entity.get_attribute('EnvizAttributes', 'TeleportOrder') || 0
            end
          
            # Update @teleports_list with the sorted entities
            @teleports_list.concat(sorted_teleports)
          
            # print_teleport_names

            # Return the updated list
            @teleports_list
        end

        # Function to print names of all teleports
        def self.print_teleport_names
            if @teleports_list.empty?
            # puts "No teleports found."
            else
            @teleports_list.each do |teleport|
                name = teleport.name
                name = "Unnamed Teleport" if name.nil? || name.empty?
                puts "Teleport Name: #{name}" if DEBUG_MODE
            end
            end
        end

        # Method to access the list of teleports from other scripts
        def self.teleports_list
            @teleports_list
        end

        class Tool_Teleport_Creator
                
            def onLButtonDown(flags, x, y, view)
                selected = Enviz_Sketchup_Helpers.raycast_hit_point(x, y, view)  # Retrieve the face using the raycast_face function
            
                if selected  # Check if a face was successfully retrieved and is not nil
                    temp_teleport = TeleportHandler.create_teleport(selected)
                    Teleports.update_teleports_list
                else
                    # puts "No face selected"  # Optional: Notify that no face was selected
                end
            end    
          
            def activate
                Sketchup::set_status_text "Click on the scene to create teleport"
                begin
                    # set_backfaces_red
                rescue Exception => e
                    puts "Failed to set backfaces: #{e.message}" if DEBUG_MODE
                end
                puts "Tool activated. Click on the scene to create teleport." if DEBUG_MODE
                Teleports.update_teleports_list
            end
          
            def deactivate(view)
                puts "Tool deactivated." if DEBUG_MODE
                view.invalidate
                Teleports.update_teleports_list
                Sketchup.active_model.select_tool(nil)
            end
            
        end

        class TeleportHandler

            def self.create_teleport(location)
                model = Sketchup.active_model
                entities = model.entities
                definition = model.definitions.add('Teleport')

                # # Switch to meters for this operation
                # original_units = model.options['UnitsOptions']['LengthUnit']
                # model.options['UnitsOptions']['LengthUnit'] = 0

                # Vert definitions
                o = 1
                o /= o.to_m
                s = o*0.1
                vertices = [

                    Geom::Point3d.new(-s, -s, s),Geom::Point3d.new(s, -s, s),Geom::Point3d.new(s, s, s),Geom::Point3d.new(-s, s, s),
                    Geom::Point3d.new(-s, -s, -s),Geom::Point3d.new(s, -s, -s),Geom::Point3d.new(s, s, -s),Geom::Point3d.new(-s, s, -s),
                    Geom::Point3d.new(s, -s, s),Geom::Point3d.new(s, s, s),Geom::Point3d.new(s, s, -s),Geom::Point3d.new(s, -s, -s),
                    Geom::Point3d.new(-s, -s, s),Geom::Point3d.new(-s, s, s),Geom::Point3d.new(-s, s, -s),Geom::Point3d.new(-s, -s, -s),
                    Geom::Point3d.new(-s, s, -s),Geom::Point3d.new(-s, s, s),Geom::Point3d.new(s, s, s),Geom::Point3d.new(s, s, -s),
                    Geom::Point3d.new(-s, -s, -s),Geom::Point3d.new(-s, -s, s),Geom::Point3d.new(s, -s, s),Geom::Point3d.new(s, -s, -s),
                    
                    Geom::Point3d.new(-s*0.5, s, -s*0.5),Geom::Point3d.new(-s*0.5, s, s*0.5),Geom::Point3d.new(-s*1.25, s*2, s*1.25),Geom::Point3d.new(-s*1.25, s*2, -s*1.25),
                    Geom::Point3d.new(s*0.5, s, s*0.5),Geom::Point3d.new(s*0.5, s, -s*0.5),Geom::Point3d.new(s*1.25, s*2, -s*1.25),Geom::Point3d.new(s*1.25, s*2, s*1.25),
                    Geom::Point3d.new(s*0.5, s, s*0.5),Geom::Point3d.new(-s*0.5, s, s*0.5),Geom::Point3d.new(-s*1.25, s*2, s*1.25),Geom::Point3d.new(s*1.25, s*2, s*1.25),
                    Geom::Point3d.new(s*0.5, s, -s*0.5),Geom::Point3d.new(-s*0.5, s, -s*0.5),Geom::Point3d.new(-s*1.25, s*2, -s*1.25),Geom::Point3d.new(s*1.25, s*2, -s*1.25),

                    Geom::Point3d.new(-s*0.1, -s*0.1, -s),Geom::Point3d.new(-s*0.1, s*0.1, -s),Geom::Point3d.new(-s*0.1, s*0.1, -o*1.5),Geom::Point3d.new(-s*0.1, -s*0.1, -o*1.5),
                    Geom::Point3d.new(s*0.1, s*0.1, -s),Geom::Point3d.new(s*0.1, -s*0.1, -s),Geom::Point3d.new(s*0.1, -s*0.1, -o*1.5),Geom::Point3d.new(s*0.1, s*0.1, -o*1.5),
                    Geom::Point3d.new(-s*0.1, s*0.1, -s),Geom::Point3d.new(s*0.1, s*0.1, -s),Geom::Point3d.new(s*0.1, s*0.1, -o*1.5),Geom::Point3d.new(-s*0.1, s*0.1, -o*1.5),
                    Geom::Point3d.new(s*0.1, -s*0.1, -s),Geom::Point3d.new(-s*0.1, -s*0.1, -s),Geom::Point3d.new(-s*0.1, -s*0.1, -o*1.5),Geom::Point3d.new(s*0.1, -s*0.1, -o*1.5)
                    
                ]

                o = 1
                o /= o.to_m
                o *= 1.5
                vertices = vertices.map do |vertex|
                    Geom::Point3d.new(vertex.x, vertex.y, vertex.z + o)
                end
                
                # Iterate over vertices in groups of four to create quads
                vertices.each_slice(4) do |v1, v2, v3, v4|
                    # Create a quad face using the current set of four vertices
                    # face = definition.entities.add_face(v1, v2, v3, v4)
                    face = definition.entities.add_face(
                        Geom::Point3d.new(v1.x, v1.y, v1.z),
                        Geom::Point3d.new(v2.x, v2.y, v2.z),
                        Geom::Point3d.new(v3.x, v3.y, v3.z),
                        Geom::Point3d.new(v4.x, v4.y, v4.z)
                    )
                    # face.pushpull(1, true) # Extrude the face (optional)
                end
                
                transformation = Geom::Transformation.new(location)
                instance = entities.add_instance(definition, transformation)
                instance.set_attribute('TeleportAttributes', 'is_teleport', true)
                disable_deletion(instance)
                lock_teleport(instance)
                add_teleport_observer(instance)
                set_teleport_attribute(instance)

                model = Sketchup.active_model
                selection = model.selection
                selection.clear
                selection.add(instance)
                instance.name = "New_Teleport"
                
                Sketchup.active_model.select_tool(nil)
                Enviz_UI.goto_teleport_creator(true)

                Teleports.update_teleports_list
                
                # puts "New teleport: #{instance.name}"
                
                instance
            end


            def self.set_teleport_attribute(component)
                model = Sketchup.active_model
                # Start an operation which can be undone in one step
                model.start_operation('Set Teleport Attribute', true)
              
                # Set an attribute on the component
                compact_teleport_orders
                new_order_num = next_available_teleport_order
                component.set_attribute('EnvizAttributes', 'TeleportType', 'EnvizTeleport')
                component.set_attribute('EnvizAttributes', 'TeleportOrder', new_order_num)
                puts "ORDER: #{new_order_num}" if DEBUG_MODE
              
                # Commit the operation
                model.commit_operation
            end     
            
            def self.next_available_teleport_order
                model = Sketchup.active_model
                entities = model.entities
              
                # Collect all TeleportOrder values
                teleport_orders = entities.map do |entity|
                  entity.get_attribute('EnvizAttributes', 'TeleportOrder')
                end.compact
              
                # Return 0 if there are no existing TeleportOrder attributes
                return 0 if teleport_orders.empty?
              
                # Sort the teleport orders
                teleport_orders.sort!
              
                # Find the first gap in the sequence
                next_available_order = 0
                teleport_orders.each_with_index do |order, index|
                  if order != index
                    next_available_order = index
                    break
                  end
                end
              
                # If no gap is found, the next available order is the next number in the sequence
                next_available_order = teleport_orders.size if next_available_order == 0
              
                next_available_order
            end

            def self.compact_teleport_orders
                model = Sketchup.active_model
                entities = model.entities
              
                # Collect all entities with the 'TeleportOrder' attribute
                teleport_entities = entities.select do |entity|
                  entity.get_attribute('EnvizAttributes', 'TeleportOrder')
                end
              
                # Return if there are no teleport entities
                return if teleport_entities.empty?
              
                # Sort entities based on the 'TeleportOrder' attribute
                sorted_entities = teleport_entities.sort_by do |entity|
                  entity.get_attribute('EnvizAttributes', 'TeleportOrder')
                end
              
                model.start_operation('Compact Teleport Orders', true)

                # Reassign 'TeleportOrder' values in a continuous sequence without gaps
                sorted_entities.each_with_index do |entity, index|
                  entity.set_attribute('EnvizAttributes', 'TeleportOrder', index)
                end
              
                # Commit the operation to make changes undoable in one step
                model.commit_operation
            end

            def self.lock_teleport(teleport)
                teleport.locked = true
            end

            def self.reorder_teleports(item_to_move, new_position)
                model = Sketchup.active_model
                entities = model.entities
            
                # Find the entity to move
                entity_to_move = entities.find do |entity|
                    entity.get_attribute('EnvizAttributes', 'TeleportOrder') == item_to_move
                end
            
                unless entity_to_move
                    puts "No entity found with TeleportOrder #{item_to_move}" if DEBUG_MODE
                    return
                end
            
                # Update the teleport list to get the current order
                Teleports.update_teleports_list
            
                # Manually find the current index of the entity to move
                current_index = nil
                ind = -1
                Teleports.teleports_list.each do |teleport|
                    ind = ind + 1
                    if teleport.get_attribute('EnvizAttributes', 'TeleportOrder') == item_to_move
                        current_index = ind
                        break
                    end
                end
            
                if current_index.nil?
                    puts "Entity with TeleportOrder #{item_to_move} not found in the list" if DEBUG_MODE
                    return
                end
            
                # Remove the entity from its current position
                Teleports.teleports_list.delete_at(current_index)
            
                # Insert the entity at the new position
                Teleports.teleports_list.insert(new_position, entity_to_move)
            
                # Update the TeleportOrder attributes
                Teleports.teleports_list.each_with_index do |entity, index|
                    entity.set_attribute('EnvizAttributes', 'TeleportOrder', index)
                end
            
                # Commit the operation to make changes undoable in one step
                model.start_operation('Reorder Teleports', true)
                model.commit_operation
            
                # Print the new order for debugging
                # Teleports.teleports_list.each do |entity|
                #     puts "Entity: #{entity}, TeleportOrder: #{entity.get_attribute('EnvizAttributes', 'TeleportOrder')}"
                # end

                compact_teleport_orders
            end


            class TeleportObserver < Sketchup::EntityObserver
                def onEraseEntity(entity)
                  return unless entity.valid?
                  allow_deletion = entity.get_attribute('TeleportAttributes', 'allow_deletion', false)
                  Sketchup.undo unless allow_deletion
                end
              
                # locking like this is probably a bad idea, so lets not do the bad thing
                # def onChangeEntity(entity)
                #   return unless entity.valid?
                #   allow_deletion = entity.get_attribute('TeleportAttributes', 'allow_deletion', false)
                #   entity.locked = true unless allow_deletion
                # end
            end          

            def self.enable_deletion(teleport)
                # Set the attribute to allow deletion
                teleport.set_attribute('TeleportAttributes', 'allow_deletion', true)
            end

            def self.disable_deletion(teleport)
                # Remove the attribute to prevent deletion
                teleport.set_attribute('TeleportAttributes', 'allow_deletion', false)
            end

            def self.cancel_deletion(teleports)
                teleports.each do |teleport|
                    disable_deletion(teleport)
                end
            end

            def self.add_teleport_observer(teleport)
                observer = TeleportObserver.new
                teleport.add_observer(observer)
            end

            def self.setup_teleport
                teleport = create_teleport
                lock_teleport(teleport)
                add_teleport_observer(teleport)
                teleport
            end

            def self.create_random_teleport
                teleport = create_teleport
                teleport.transformation = Geom::Transformation.new([rand * 2 - 1, rand * 2 - 1, rand * 2 - 1].map { |coord| coord * 39.3701 })
                lock_teleport(teleport)
                add_teleport_observer(teleport)
                teleport
            end

            def self.delete_teleport(order)
                model = Sketchup.active_model
                entities = model.active_entities
              
                # Start a new operation which allows undoing all steps in one action
                model.start_operation("Delete Teleport", true)
              
                # Find the entity with the matching TeleportOrder
                entity_to_delete = entities.find do |entity|
                  entity.get_attribute('EnvizAttributes', 'TeleportOrder') == order
                end
              
                if entity_to_delete
                  # Allow deletion by setting the attribute
                  enable_deletion(entity_to_delete)
                  entity_to_delete.locked = false
                  remove_observers_from_teleport(entity_to_delete)
                  puts "Preparing to delete teleport: #{entity_to_delete}, allow_deletion: #{entity_to_delete.get_attribute('TeleportAttributes', 'allow_deletion', false)}" if DEBUG_MODE
              
                  # Delete the teleport entity
                  deleted_count = entities.erase_entities([entity_to_delete])
                  puts "Deleted #{deleted_count} teleport(s)." if DEBUG_MODE
                else
                  puts "No teleport found with TeleportOrder #{order}." if DEBUG_MODE
                end
              
                # Commit the operation to allow for undo
                model.commit_operation
              
                # Update the teleport list if the method is defined
                Teleports.update_teleports_list if defined?(Teleports.update_teleports_list)
                compact_teleport_orders
            end

            def self.get_current_selection
                Sketchup.active_model.selection.to_a
            end

            def self.activate_delete_teleport
                selected_teleports = get_current_selection
                if selected_teleports.empty?
                    puts "No objects are selected." if DEBUG_MODE
                else
                    delete_teleport(selected_teleports)
                end
            end

            def self.remove_observers_from_teleport(entity)
                # Check if the entity is valid and supports observers
                if entity.valid? && entity.respond_to?(:observers)
                    observers = entity.observers
                    observers.each do |observer|
                        entity.remove_observer(observer)
                        puts "Removed observer: #{observer.class}" if DEBUG_MODE
                    end
                else

                end
            end

        end

    end
end