require 'sketchup'
Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_small_tools')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
Sketchup::require('enviz_sketchup_plugin/enviz_logging')
Sketchup::require('enviz_sketchup_plugin/teleport_handler')
Sketchup::require('enviz_sketchup_plugin/tool_flipper')
Sketchup::require('enviz_sketchup_plugin/tool_dollhouse_creator')
require 'json'
require 'zlib'
require 'fileutils'

module Enviz_SketchUp_Plugin
    module Exporter

        include Teleports
        include Dollhouse 
        include Validator

        def self.version_to_int(version)
            parts = version.split('.').map(&:to_i)
            val = parts[0] * 100 + parts[1] * 10 + parts[2]
            puts "v#{val}" if DEBUG_MODE
            val
        end

        def self.gzip_directory(directory_path, output_file)

            Zlib::GzipWriter.open(output_file) do |gz|

                gz.write("enviz")
                gz.write(Enviz_SketchUp_Plugin::VERSION)

                Dir[File.join(directory_path, '**', '*')].each do |file|
           
                    next if File.directory?(file)

                    file_size = File.size(file)

                    # Write file path and its size for index
                    gz.write "file:#{file}\nsize:#{file_size}\n"
            
                    # Read and write file content
                    File.open(file, 'rb') do |input|

                        while chunk = input.read(16 * 1024) do

                            gz.write(chunk)

                        end

                    end
          
                    # File separator
                    gz.write "\n---end-of-file-enviz---\n"
           
                end

            end
            
        end

        def self.decompress_gz_file(input_file, output_directory)
           
            Zlib::GzipReader.open(input_file) do |gz|

                content = gz.read
                # Split the binary content using the binary version of the delimiter
                files_data = content.force_encoding('binary').split("\n---end-of-file-enviz---\n".encode('binary'))

                files_data.each do |file_data|

                next if file_data.strip.empty?

                    # Force binary encoding to handle non-text files correctly
                    file_data.force_encoding('binary')

                    # Extract the file path and size from the first line which starts with "file: "
                    header, *file_contents = file_data.split("\n".encode('binary'))
                    header_info = header.strip.match(/^file: (?<path>.+) size: (?<size>\d+)$/)
                    puts header if DEBUG_MODE
                    puts header_info if DEBUG_MODE
                    next unless header_info # Skip if the header is not correctly formatted

                    file_path = header_info[:path]
                    file_size = header_info[:size].to_i
                    file_contents = file_contents.join("\n".encode('binary'))

                    # Truncate the file_contents to the size specified in the header
                    file_contents = file_contents[0...file_size]

                    # Ensure output directory exists and write file in binary mode
                    full_path = File.join(output_directory, File.basename(file_path))
                    FileUtils.mkdir_p(File.dirname(full_path))
                    File.open(full_path, 'wb') do |file|
                        file.write(file_contents)
                    end

                end

            end

        end

        def self.export_model

            model = Sketchup.active_model
            modelPath = model.path
            if model.modified? or !modelPath or modelPath.empty? then
                UI.messagebox("You must save your SketchUp model before export", MB_OK)
                return
            else

                Validator.validate_model

                if Validator.is_valid

                    # set internal file units to meters, 'units' param of export options doesn't appear to do anything
                    # this seems to be the only way to make files export the way we want
                    model = Sketchup.active_model
                    model.options["UnitsOptions"]["LengthFormat"] = 0
                    model.options["UnitsOptions"]["LengthUnit"] = 3
                    model.options["UnitsOptions"]["AreaUnit"] = 3
                    model.options["UnitsOptions"]["VolumeUnit"] = 3
                
                    model = Sketchup.active_model

                    path = UI.select_directory(title: "Select Export Directory")
                    if path  # false if user cancelled dialog
                        
                        temp_path = File.join(path, "enviz_temp")
                        
                        options_fbx = { 
                            :show_summary => false,
                            :units => "model", 
                            :triangulated_faces => true,
                            :doublesided_faces => false,
                            :texture_maps => true,
                            :seperate_disconnected_faces => false,
                            :swap_yz => true,
                            :selectionset_only => false
                        }

                        # Create a subdirectory 'enviz_temp' within the chosen directory
                        FileUtils.mkdir_p(temp_path) unless Dir.exist?(temp_path)
                        
                        if modelPath.index('\\') then
                            modelPath = modelPath.split("\\").last
                        else
                            modelPath = modelPath.split("/").last
                        end

                        hide_teleports(true)
                        
                        file_name = modelPath.split(".")[0] + ".fbx"
                        export_path = File.join(temp_path, file_name)
                        model.export(export_path, options_fbx)

                        hide_teleports(false)

                        # JSON
                        data = gather_data
                        json_string = JSON.pretty_generate(data)
            
                        file_path = File.join(temp_path, "enviz_data.json")
                        
                        File.open(file_path, "w") do |file|
                            file.write(json_string)
                        end
                        puts "Written Enviz JSON" if DEBUG_MODE

                        puts "Writing logs" if DEBUG_MODE
                        
                        current_time = Time.now.strftime("%Y-%m-%d_%H-%M-%S")
                        file_dir = "#{temp_path}/log_#{current_time}.txt"
                        Enviz_Logging.write_logs_to_file(file_dir)
                        puts "#{file_dir}" if DEBUG_MODE

                        puts "Written logs" if DEBUG_MODE
            
                        zip_dir = "#{path}/upload.enviz"
                        gzip_directory(temp_path, zip_dir)
                        # decompress_gz_file(zip_dir, temp_path)
            
                        puts "Compressed output - #{zip_dir}" if DEBUG_MODE
            
                        delete_directory_and_contents(temp_path)
            
                        puts "File cleanup complete" if DEBUG_MODE

                    end

                
                else

                    puts "Model invalid" if DEBUG_MODE
                    Enviz_UI.goto_main(true)

                end

            end

        end

        def self.delete_directory_and_contents(path)

            FileUtils.rm_rf(path)
        
        end

        def self.hide_teleports(should_hide)

            Teleports.update_teleports_list
            teleport_list = Teleports.teleports_list

            teleport_list.each do |teleport|

                teleport.hidden = should_hide

            end

        end

        def self.hide_empties(should_hide)

            model = Sketchup.active_model
            entities = model.entities

            entities.each do |entity|
               
               if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
               
                    if entity.entities.empty?
               
                        entity.hidden = should_hide
               
                    end
               
                end
            
            end

        end

        def self.gather_data

            Teleports.update_teleports_list
            teleport_list = Teleports.teleports_list
        
            puts "TELEPORTS - #{teleport_list.length}" if DEBUG_MODE
        
            dp = Dollhouse.get_dollhouse_pivot
            dv = Dollhouse.get_dollhouse_view
        
            # Prepare data with dollhouse pivot and view
            data = {
                "plugin_version" => "#{Enviz_SketchUp_Plugin::VERSION}",
                "dollhouse_pivot" => "#{dp.x.to_m},#{dp.y.to_m},#{dp.z.to_m}",
                "dollhouse_view" => "#{dv.x.to_m},#{dv.y.to_m},#{dv.z.to_m}",
                "teleports" => []
            }
        
            # Sort teleport_list by the 'TeleportOrder' attribute
            sorted_teleport_list = teleport_list.sort_by do |teleport|
                teleport.get_attribute('EnvizAttributes', 'TeleportOrder')
            end

            # Add each teleport's position to the data structure
            sorted_teleport_list.each_with_index do |teleport, index|
                pos = teleport.transformation.origin
                rot = (angle_in_plane(teleport.transformation.xaxis, X_AXIS) * 180 / Math::PI).to_i
                data["teleports"] << "#{teleport.name}[#{index}],#{pos.x.to_m},#{pos.y.to_m},#{pos.z.to_m},#{rot}"
            end
        
            return data

        end

        def self.angle_in_plane(vector1, vector2, normal = Z_AXIS)
            Math.atan2((vector2 * vector1) % normal, vector1 % vector2)
        end

    end
end