require 'sketchup'
Sketchup::require('enviz_sketchup_plugin/enviz_sketchup_helpers')
Sketchup::require('enviz_sketchup_plugin/enviz_small_tools')
Sketchup::require('enviz_sketchup_plugin/enviz_ui')
Sketchup::require('enviz_sketchup_plugin/teleport_handler')
Sketchup::require('enviz_sketchup_plugin/tool_flipper')
Sketchup::require('enviz_sketchup_plugin/tool_dollhouse_creator')

module Enviz_SketchUp_Plugin
    module Validator

        include Teleports
        include Dollhouse

        MAX_POLYS = 5000000
        MAX_SURFACE = 40000
        MAX_OBJECTS = 10000
        MAX_TEXTURES = 1000

        # Initialize the class instance variables
        @is_valid = false
        @poly_pass = false
        @surface_pass = false
        @object_pass = false
        @texture_pass = false
        @poly_string = ""
        @surface_string = ""
        @object_string = ""
        @texture_string = ""

        # Method to set the is_valid variable
        def self.is_valid=(value)
            @is_valid = value
        end
        def self.is_valid
            @is_valid
        end
        def self.invalidate
            @is_valid = false
        end

        def self.poly_pass=(value)
            @poly_pass = value
        end
        def self.poly_pass
            @poly_pass
        end

        def self.poly_string=(value)
            @poly_string = value
        end
        def self.poly_string
            @poly_string
        end

        def self.surface_pass=(value)
            @surface_pass = value
        end
        def self.surface_pass
            @surface_pass
        end

        def self.surface_string=(value)
            @surface_string = value
        end
        def self.surface_string
            @surface_string
        end

        def self.object_string=(value)
            @object_string = value
        end
        def self.object_string
            @object_string
        end
        def self.object_pass=(value)
            @object_pass = value
        end
        def self.object_pass
            @object_pass
        end

        def self.texture_string=(value)
            @texture_string = value
        end
        def self.texture_string
            @texture_string
        end
        def self.texture_pass=(value)
            @texture_pass = value
        end
        def self.texture_pass
            @texture_pass
        end

        def self.validate_model

            # teleports placed
            Teleports.update_teleports_list
            teleports_placed = Teleports.teleports_list.length
            puts "TELEPORTS - #{teleports_placed}" if DEBUG_MODE

            # dollhouse set up
            dollhouse_set = Dollhouse.is_dollhouse_set
            puts "DOLLHOUSE SET - #{dollhouse_set}" if DEBUG_MODE

            # number of polys
            poly_pass = false
            validate_polys

            # total surface area
            surface_pass = false
            validate_surface_area
            
            # validate objects
            object_pass = false
            validate_objects

            # validate textures
            texture_pass = false
            validate_textures

            if teleports_placed > 0 && dollhouse_set && @poly_pass && @surface_pass && @object_pass && @texture_pass
                @is_valid = true
            else
                @is_valid = false
            end

        end

        def self.validate_objects
            model = Sketchup.active_model
            entities = model.entities
        
            # Initialize a stack with top-level entities
            stack = entities.to_a
            mesh_object_count = 0
        
            # Iterate through the stack
            until stack.empty?
                entity = stack.pop
        
                if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                    mesh_object_count += 1
                    # Add nested entities to the stack
                    stack.concat(entity.definition.entities.to_a)
                end
            end
        
            # Set object_pass based on the count
            @object_pass = mesh_object_count <= MAX_OBJECTS
        
            mesh_object_count_s = format_number(mesh_object_count)
            max_objects_s = format_number(MAX_OBJECTS)
            concat_string = "(#{mesh_object_count_s}/#{max_objects_s})"
            @object_string = concat_string
        
            puts "OBJECTS - #{@object_string}" if DEBUG_MODE
        end

        def self.validate_textures
            model = Sketchup.active_model
            entities = model.entities
        
            used_materials = []
        
            # Collect materials used by faces and groups/components in the model
            entities.each do |entity|
                if entity.is_a?(Sketchup::Face)
                    material = entity.material
                    used_materials << material if material && material.texture
        
                    back_material = entity.back_material
                    used_materials << back_material if back_material && back_material.texture
                elsif entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                    entity.definition.entities.each do |sub_entity|
                        if sub_entity.is_a?(Sketchup::Face)
                            material = sub_entity.material
                            used_materials << material if material && material.texture
        
                            back_material = sub_entity.back_material
                            used_materials << back_material if back_material && back_material.texture
                        end
                    end
                end
            end
        
            # Remove duplicates
            used_materials.uniq!
        
            # Count the number of materials with textures
            texture_count = used_materials.count
        
            # Set texture_pass based on the count
            @texture_pass = texture_count <= MAX_TEXTURES
        
            texture_count_s = format_number(texture_count)
            max_textures_s = format_number(MAX_TEXTURES)
            concat_string = "(#{texture_count_s}/#{max_textures_s})"
            @texture_string = concat_string
        
            puts "TEXTURES - #{@texture_string}" if DEBUG_MODE
        end

        def self.validate_polys
            polys = count_triangles
            @poly_pass = polys <= MAX_POLYS

            polys_s = format_number(polys)
            max_polys_s = format_number(MAX_POLYS)
            concat_string = "(#{polys_s}/#{max_polys_s})"
            @poly_string = concat_string

            puts "POLYS - #{@poly_string}" if DEBUG_MODE
        end
      
        def self.validate_surface_area
            surface = count_surface_area
            @surface_pass = surface <= MAX_SURFACE

            surface_s = format_number(surface)
            max_surface_s = format_number(MAX_SURFACE)
            concat_string = "(#{surface_s}/#{max_surface_s})"
            @surface_string = concat_string

            puts "SURFACE - #{@surface_string}" if DEBUG_MODE
        end

        def self.format_number(number)
            case number
            when 0...1000
                number.to_i.to_s
            when 1000...1_000_000
                formatted = (number / 1000.0).round(1).to_s
                formatted.sub('.0', '') + 'k'  # Remove '.0' if it occurs
            else
                formatted = (number / 1_000_000.0).round(2).to_s
                formatted.sub('.00', '') + 'mil'  # Remove '.00' if it occurs
            end
        end    

        def self.count_triangles
            model = Sketchup.active_model
            entities = model.entities
            total_triangles = [0]  # Initialize as an array with one element
        
            # Function to recursively count triangles in groups or component instances
            def self.traverse_entity(entity, triangles)
                if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                    entity.definition.entities.each { |sub_entity| traverse_entity(sub_entity, triangles) }
                elsif entity.is_a?(Sketchup::Face)
                    # Each face can be divided into triangles
                    triangles_count = entity.mesh.polygons.size
                    triangles[0] += triangles_count  # Increment the count in the array
                end
            end
        
            entities.each do |entity|
                traverse_entity(entity, total_triangles)  # Pass the array
            end
        
            total_triangles[0]  # Return the count from the array
        end   

        def self.count_surface_area
            model = Sketchup.active_model
            entities = model.entities
            total_surface_area = 0.0
        
            # Function to recursively calculate surface area of faces, considering transformations
            def self.traverse_entity(entity, surface_area, transformation = IDENTITY)
                if entity.is_a?(Sketchup::Group) || entity.is_a?(Sketchup::ComponentInstance)
                    # Apply the current entity's transformation to the accumulated transformation
                    new_transformation = transformation * entity.transformation
                    entity.definition.entities.each { |sub_entity| traverse_entity(sub_entity, surface_area, new_transformation) }
                elsif entity.is_a?(Sketchup::Face)
                    # Calculate the transformed area of the face
                    transformed_area = entity.area(transform=transformation)
                    # SketchUp API returns area in square inches, convert it to square meters if required
                    area_meters = transformed_area.to_m.to_m  # Conversion factor from square inches to square meters
                    surface_area[0] += area_meters
                end
            end
        
            # Use an array to pass the running total by reference
            total_area = [0.0]
            entities.each do |entity|
                traverse_entity(entity, total_area)
            end
            total_surface_area = total_area[0]
        
            total_surface_area
        end
    end
end